
#include <sol.h>
#include <iostream>
#include <fstream>

using namespace std;

int bufsize = 10000;
int **duration, nMachines, nJobs, lb, opt;

int maxfsble;
int minfsble;
int makespan;


void readData( char* filename )
{
  char buf[bufsize];
  buf[0] = '#';
  ifstream infile( filename, ios_base::in );
  while( buf[0] == '#' )
    {
      infile.getline( buf, bufsize, '\n' );
    }
  int i=0; 
  while( buf[i] != ' ' ) ++i;
  buf[i] = '\0';
  lb = atoi( buf );
  i+=2;
  int j = i;
  while( buf[i] != ' ' && buf[i] != '\n' && buf[i] != '\0' ) ++i;
  buf[i] = '\0';
  opt = atoi( &(buf[j]) );
  infile.getline( buf, bufsize, '\n' );
  
  infile >> nJobs;
  infile >> nMachines;

  duration = new int*[nJobs];
  for(i=0; i<nJobs; ++i)
    duration[i] = new int[nMachines];

  infile.getline( buf, bufsize, '\n' );
  infile.getline( buf, bufsize, '\n' );

  for(i=0; i<nJobs; ++i) 
    for(j=0; j<nMachines; ++j) 
      infile >> duration[i][j];
}


int upperbound() {
  int i, j, mkp=0, task, boundJob[nJobs], boundMachine[nJobs]; 
  std::fill(boundJob, boundJob+nJobs, 0);
  std::fill(boundMachine, boundMachine+nJobs, 0);

  assert( nJobs == nMachines );

  for(i=0; i<nJobs; ++i)
    {
      for(j=0; j<nJobs; ++j)
	{
	  task = std::max( boundJob[j], boundMachine[(i+j) % nJobs] );

	  //cout << task << " (" << j << "," << ((i+j) % nJobs) << ") \t";

	  task += duration[j][(i+j) % nJobs];
	  boundJob[j] = task;
	  boundMachine[(i+j) % nJobs] = task;
	  if( task > mkp ) mkp = task;
	}

      //      cout << endl;
    }
  return mkp;
}


int lowerbound()
{
  int i, j, mkp=0, boundJob[nJobs], boundMachine[nMachines]; 
  std::fill(boundJob, boundJob+nJobs, 0);
  std::fill(boundMachine, boundMachine+nMachines, 0);

  for(i=0; i<nJobs; ++i) {
    for(j=0; j<nMachines; ++j)
      {
	boundJob[i] += duration[i][j];
	boundMachine[j] += duration[i][j];
      }
  }

  for(i=0; i<nJobs; ++i)
    if( mkp < boundJob[i] ) mkp = boundJob[i];
  for(i=0; i<nMachines; ++i)
    if( mkp < boundMachine[i] ) mkp = boundMachine[i];

  return mkp;
}

void setup(CSP& model, VarArray& disjuncts, VarArray& tasks) {
  int i,j,k;
  for(i=0; i<nJobs; ++i) 
    for(j=0; j<nMachines; ++j) 
      tasks.add( Variable(0, makespan-duration[i][j]) );

  for(i=0; i<nMachines; ++i) {
    for(j=0; j<nJobs; ++j) 
      for(k=j+1; k<nJobs; ++k) 
	disjuncts.add( Disjunctive( tasks[j*nMachines+i], duration[j][i],
				    tasks[k*nMachines+i], duration[k][i] ) );
  }

  for(i=0; i<nJobs; ++i) {
    for(j=0; j<nMachines; ++j) 
      for(k=j+1; k<nMachines; ++k) 
	disjuncts.add( Disjunctive( tasks[i*nMachines+j], duration[i][j],
				    tasks[i*nMachines+k], duration[i][k] ) );
  }

  model.add( disjuncts );

}


int main( int argc, char** argv )
{
  if( argc < 2 )
    {
      cerr << "need a data file" << endl;
      exit( 0 );
    }

  int cutoff = ( argc > 2 ? atoi(argv[2]) : 5 );

  int i, j;
  readData( argv[1] );

  maxfsble = upperbound();
  minfsble = lowerbound();

  cout << minfsble << " <= " << lb << " <= " << opt << " <= " << maxfsble << endl;
  
  // Model

  bool solved = true;
  while( minfsble < maxfsble ) {
    makespan = (int)(floor(((double)minfsble + (double)maxfsble)/2));
      
    cout << "\t" << setw(5) << minfsble << " " << setw(5) << maxfsble 
	 << " (" << setw(5) << makespan << "):\t";
    cout.flush();
                      
    VarArray disjuncts;
    VarArray tasks;
    CSP model;
    setup(model, disjuncts, tasks);

    DomOverWDeg heuristic;
    Solver s(model, disjuncts, heuristic);
    s.setTimeLimit(cutoff);


    //    s.print( cout );
    //    cout << endl;
    //    exit( 0 );
      
    s.setRandomized();
    int result = s.solve_and_restart(Solver::GEOMETRIC, 64, 1.2);
    //int result = s.solve();

    if( result == Solver::SAT ) {
      maxfsble = makespan;
    } else {
      minfsble = makespan+1;
      if( result == Solver::LIMITOUT )
	solved = false;
    }
    
    s.printStatistics(std::cout);
    cout << endl;
  }

  if( !solved ) {
    VarArray disjuncts;
    VarArray tasks;
    VarArray endlasttask( nJobs*nMachines, 0, maxfsble );

    CSP model;
    makespan = maxfsble-1;
    setup(model, disjuncts, tasks);
    for(i=0; i<nJobs; ++i)
      for(j=0; j<nMachines; ++j) 
	model.add( tasks[i*nMachines+j] + duration[i][j] == endlasttask[i*nMachines+j] );

    model.add( Minimise(Max(endlasttask)) );

    DomOverWDeg heuristic;
    Solver s(model, disjuncts, heuristic);
    s.solve();
    maxfsble = s.goal->upper_bound;
  }

  cout << maxfsble << " " << opt << endl;


}



  


  




